@@ -2,11 +2,14 @@ package Mojo::SMTP::Client;
 
 use Mojo::Base 'Mojo::EventEmitter';
 use Mojo::IOLoop;
+use Mojo::IOLoop::Client;
 use Mojo::IOLoop::Delay;
+use Mojo::IOLoop::Stream;
 use Mojo::SMTP::Client::Exception;
 use Carp;
+use Scalar::Util 'weaken';
 
-our $VERSION = 0.03;
+our $VERSION = 0.04;
 
 use constant {
 	CMD_OK       => 2,
@@ -62,9 +65,12 @@ sub send {
 	my $expected_code;
 	my $nb = $cb ? 1 : 0;
 	
+	$self->{finished} = 0;
+	weaken(my $this = $self);
+	
 	my $resp_checker = sub {
 		my ($delay, $resp) = @_;
-		$self->emit(response => $self->{last_cmd}, $resp);
+		$this->emit(response => $this->{last_cmd}, $resp);
 		
 		die $resp->{error} if $resp->{error};
 		substr($resp->{code}, 0, 1) == $expected_code
@@ -72,30 +78,58 @@ sub send {
 		$delay->pass($resp);
 	};
 	
-	delete($self->{stream})->close
-		if $self->{stream} && $self->{server} ne $self->_server;
+	# user changed SMTP server or server sent smth while it shouldn't
+	if ($this->{stream} && (($this->{server} ne $this->_server) ||
+	     ($this->{stream}->is_readable && grep {$this->{last_cmd} == $_} (CMD_CONNECT, CMD_DATA_END, CMD_RESET)))
+	) {
+		$this->_rm_stream();
+	}
 	
-	unless ($self->{stream}) {
+	unless ($this->{stream}) {
 		push @steps, sub {
 			my $delay = shift;
 			# connect
-			$self->emit('start');
-			$self->{server} = $self->_server;
-			$self->{last_cmd} = CMD_CONNECT;
-			$self->_ioloop($nb)->client(
-				address => $self->address,
-				port    => $self->port,
-				timeout => $self->connect_timeout,
-				$delay->begin
-			)
+			$this->emit('start');
+			$this->{server} = $this->_server;
+			$this->{last_cmd} = CMD_CONNECT;
+			
+			my $connect_cb = $delay->begin;
+			$this->{client} = Mojo::IOLoop::Client->new(reactor => $this->_ioloop($nb)->reactor);
+			$this->{client}->on(connect => $connect_cb);
+			$this->{client}->on(error => $connect_cb);
+			$this->{client}->connect(
+				address => $this->address,
+				port    => $this->port,
+				timeout => $this->connect_timeout
+			);
 		},
 		sub {
 			# read response
-			my ($delay, $err, $stream) = @_;
-			Mojo::SMTP::Client::Exception::Stream->throw($err) if $err;
+			my $delay = shift;
+			delete $this->{client};
+			# check is this a handle
+			Mojo::SMTP::Client::Exception::Stream->throw($_[0]) unless eval { *{$_[0]} };
 			
-			$self->{stream} = $stream;
-			$self->_read_response($delay->begin);
+			my $error_handler = sub {
+				delete($this->{cleanup_cb})->() if $this->{cleanup_cb};
+				$this->{delay}->remaining([]); # remove remaining steps
+				$this->_rm_stream();
+				$this->{delay}->emit(finish => {error => $_[0]});
+			};
+			
+			$this->{stream} = Mojo::IOLoop::Stream->new($_[0]);
+			$this->{stream}->reactor($this->_ioloop($nb)->reactor);
+			$this->{stream}->start;
+			$this->{stream}->on(timeout => sub {
+				$error_handler->(Mojo::SMTP::Client::Exception::Stream->new('Inactivity timeout'));
+			});
+			$this->{stream}->on(error => sub {
+				$error_handler->(Mojo::SMTP::Client::Exception::Stream->new($_[-1]));
+			});
+			$this->{stream}->on(close => sub {
+				$error_handler->(Mojo::SMTP::Client::Exception::Stream->new('Socket closed unexpectedly by remote side'));
+			});
+			$this->_read_response($delay->begin, 0);
 			$expected_code = CMD_OK;
 		},
 		# check response
@@ -103,8 +137,8 @@ sub send {
 		# HELO
 		sub {
 			my $delay = shift;
-			$self->_cmd('EHLO ' . $self->hello, CMD_EHLO);
-			$self->_read_response($delay->begin);
+			$this->_cmd('EHLO ' . $this->hello, CMD_EHLO);
+			$this->_read_response($delay->begin, 0);
 			$expected_code = CMD_OK;
 		}, 
 		sub {
@@ -112,8 +146,8 @@ sub send {
 			if (my $err = $@) {
 				die $err unless $err->isa('Mojo::SMTP::Client::Exception::Response');
 				my $delay = shift;
-				$self->_cmd('HELO ' . $self->hello, CMD_HELO);
-				$self->_read_response($delay->begin);
+				$this->_cmd('HELO ' . $this->hello, CMD_HELO);
+				$this->_read_response($delay->begin, 0);
 			}
 		},
 		sub {
@@ -123,7 +157,7 @@ sub send {
 		};
 	}
 	else {
-		$self->{stream}->start;
+		$this->{stream}->start;
 	}
 	
 	for (my $i=0; $i<@cmd; $i+=2) {
@@ -132,18 +166,22 @@ sub send {
 		if ($cmd[$i] eq 'from') { # FROM
 			push @steps, sub {
 				my $delay = shift;
-				$self->_cmd('MAIL FROM:<'.$cmd[$mi].'>', CMD_FROM);
-				$self->_read_response($delay->begin);
+				$this->_cmd('MAIL FROM:<'.$cmd[$mi].'>', CMD_FROM);
+				$this->_read_response($delay->begin, $mi == $#cmd);
 				$expected_code = CMD_OK;
 			},
 			$resp_checker
 		}
 		elsif ($cmd[$i] eq 'to') { # TO
+			my $cur = 0;
+			my $count = ref $cmd[$mi] ? @{$cmd[$mi]} : 1;
+			
 			for my $to (ref $cmd[$mi] ? @{$cmd[$mi]} : $cmd[$mi]) {
+				my $j = ++$cur;
 				push @steps, sub {
 					my $delay = shift;
-					$self->_cmd('RCPT TO:<'.$to.'>', CMD_TO);
-					$self->_read_response($delay->begin);
+					$this->_cmd('RCPT TO:<'.$to.'>', CMD_TO);
+					$this->_read_response($delay->begin, $mi == $#cmd && $j == $count);
 					$expected_code = CMD_OK;
 				},
 				$resp_checker
@@ -153,8 +191,8 @@ sub send {
 			# DATA
 			push @steps, sub {
 				my $delay = shift;
-				$self->_cmd('DATA', CMD_DATA);
-				$self->_read_response($delay->begin);
+				$this->_cmd('DATA', CMD_DATA);
+				$this->_read_response($delay->begin, 0);
 				$expected_code = CMD_MORE;
 			},
 			$resp_checker;
@@ -167,24 +205,22 @@ sub send {
 					my $delay = shift;
 					unless ($data_writer_cb) {
 						$data_writer_cb = $delay->begin;
-						$self->_set_errors_handler(sub {
-							$data_writer_cb->($delay, @_);
+						$this->{cleanup_cb} = sub {
 							undef $data_writer;
-						});
+						};
 					}
 					
 					my $data = $cmd[$mi]->();
 					
 					unless (length(ref $data ? $$data : $data) > 0) {
-						$self->_cmd(($was_nl ? '' : CRLF).'.', CMD_DATA_END);
-						$self->_read_response($data_writer_cb);
-						$self->_set_errors_handler(undef);
+						$this->_cmd(($was_nl ? '' : CRLF).'.', CMD_DATA_END);
+						$this->_read_response($data_writer_cb, $mi == $#cmd);
 						$expected_code = CMD_OK;
-						return undef $data_writer;
+						return delete($this->{cleanup_cb})->();
 					}
 					
 					$was_nl = _has_nl($data);
-					$self->{stream}->write(ref $data ? $$data : $data, $data_writer);
+					$this->{stream}->write(ref $data ? $$data : $data, $data_writer);
 				};
 				
 				push @steps, $data_writer, $resp_checker;
@@ -192,21 +228,12 @@ sub send {
 			else {
 				push @steps, sub {
 					my $delay = shift;
-					my $data_writer_cb = $delay->begin;
-					$self->{stream}->write(ref $cmd[$mi] ? ${$cmd[$mi]} : $cmd[$mi], $data_writer_cb);
-					$self->_set_errors_handler(sub {
-						$data_writer_cb->(@_);
-					});
+					$this->{stream}->write(ref $cmd[$mi] ? ${$cmd[$mi]} : $cmd[$mi], $delay->begin);
 				},
 				sub {
-					my ($delay, $resp) = @_;
-					if ($resp && $resp->{error}) {
-						die $resp->{error};
-					}
-					
-					$self->_set_errors_handler(undef);
-					$self->_cmd((_has_nl($cmd[$mi]) ? '' : CRLF).'.', CMD_DATA_END);
-					$self->_read_response($delay->begin);
+					my $delay = shift;
+					$this->_cmd((_has_nl($cmd[$mi]) ? '' : CRLF).'.', CMD_DATA_END);
+					$this->_read_response($delay->begin, $mi == $#cmd);
 					$expected_code = CMD_OK;
 				},
 				$resp_checker
@@ -215,8 +242,8 @@ sub send {
 		elsif ($cmd[$i] eq 'reset') { # RESET
 			push @steps, sub {
 				my $delay = shift;
-				$self->_cmd('RSET', CMD_RESET);
-				$self->_read_response($delay->begin);
+				$this->_cmd('RSET', CMD_RESET);
+				$this->_read_response($delay->begin, $mi == $#cmd);
 				$expected_code = CMD_OK;
 			},
 			$resp_checker
@@ -224,13 +251,13 @@ sub send {
 		elsif ($cmd[$i] eq 'quit') { # QUIT
 			push @steps, sub {
 				my $delay = shift;
-				$self->_cmd('QUIT', CMD_QUIT);
-				$self->_read_response($delay->begin);
+				$this->_cmd('QUIT', CMD_QUIT);
+				$this->_read_response($delay->begin, $mi == $#cmd);
 				$expected_code = CMD_OK;
 			},
 			$resp_checker, sub {
 				my $delay = shift;
-				delete($self->{stream})->close;
+				$this->_rm_stream();
 				$delay->pass(@_);
 			}
 		}
@@ -240,15 +267,15 @@ sub send {
 	}
 	
 	# non-blocking
-	my $delay = Mojo::IOLoop::Delay->new(ioloop => $self->_ioloop($nb))->steps(@steps)->catch(sub {
+	my $delay = $this->{delay} = Mojo::IOLoop::Delay->new(ioloop => $this->_ioloop($nb))->steps(@steps)->catch(sub {
 		shift->emit(finish => {error => $_[0]});
 	});
 	$delay->on(finish => sub {
-		if ($self->{stream}) {
-			$self->{stream}->timeout(0);
-			$self->{stream}->stop;
+		if ($cb) {
+			$cb->($self, $_[1]);
+			$cb = undef;
+			delete $self->{delay};
 		}
-		$cb->($self, $_[1]);
 	});
 	
 	# blocking
@@ -272,47 +299,34 @@ sub _server {
 	return $self->address.':'.$self->port;
 }
 
-sub _set_errors_handler {
-	my ($self, $cb) = @_;
-	
-	unless ($cb) {
-		return
-			$self->{stream}->unsubscribe('timeout')
-			               ->unsubscribe('error')
-			               ->unsubscribe('close');
-	}
-	
-	$self->{stream}->on(timeout => sub {
-		delete($self->{stream});
-		$cb->($self, {error => Mojo::SMTP::Client::Exception::Stream->new('Inactivity timeout')});
-	});
-	$self->{stream}->on(error => sub {
-		delete($self->{stream});
-		$cb->($self, {error => Mojo::SMTP::Client::Exception::Stream->new($_[-1])});
-	});
-	$self->{stream}->on(close => sub {
-		delete($self->{stream});
-		$cb->($self, {error => Mojo::SMTP::Client::Exception::Stream->new('Socket closed unexpectedly by remote side')});
-	});
-}
-
 sub _cmd {
 	my ($self, $cmd, $cmd_const) = @_;
 	$self->{last_cmd} = $cmd_const;
 	$self->{stream}->write($cmd.CRLF);
 }
 
+sub _rm_stream {
+	my $self = shift;
+	$self->{stream}->unsubscribe('close')
+	               ->unsubscribe('timeout')
+	               ->unsubscribe('error')
+	               ->unsubscribe('read');
+	delete $self->{stream};
+}
+
 sub _read_response {
-	my ($self, $cb) = @_;
+	my ($self, $cb, $last) = @_;
 	$self->{stream}->timeout($self->inactivity_timeout);
-	$self->_set_errors_handler($cb);
 	my $resp = '';
 	
 	$self->{stream}->on(read => sub {
 		$resp .= $_[-1];
 		if ($resp =~ /^\d+(?:\s[^\n]*)?\n$/m) {
 			$self->{stream}->unsubscribe('read');
-			$self->_set_errors_handler(undef);
+			if ($last) {
+				$self->{stream}->timeout(0);
+				$self->{stream}->stop;
+			}
 			$cb->($self, _parse_response($resp));
 		}
 	});
@@ -340,6 +354,13 @@ sub _has_nl {
 	$_[0] =~ /\012$/;
 }
 
+sub DESTROY {
+	my $self = shift;
+	if ($self->{stream}) {
+		$self->_rm_stream();
+	}
+}
+
 1;
 
 __END__
@@ -437,11 +458,11 @@ with object method call
 
 =head2 address
 
-Address for SMTP server (ip or domain name). Default is C<localhost>
+Address of SMTP server (ip or domain name). Default is C<localhost>
 
 =head2 port
 
-Port for SMTP server. Default is C<25>
+Port of SMTP server. Default is C<25>
 
 =head2 hello
 
@@ -513,7 +534,7 @@ Note: transaction considered started after C<MAIL FROM> (C<from>) command.
 =item data
 
 Email body to be sent. Value for this command should be a string (or reference to a string) with email body or reference to subroutine
-each call of which should return some chunk of the email as string (or reference to a string) and empty string (or reference to a string)
+each call of which should return some chunk of the email as string (or reference to a string) and empty string (or reference to empty string)
 at the end (useful to send big emails in memory-efficient way)
 
 	$smtp->send(data => "Subject: This is my first message\r\n\r\nSent from Mojolicious app");
@@ -686,8 +707,8 @@ as argument, so it will be super easy to send our big email in memory-efficient
 
 Sometimes it is more suitable to send message directly to SMTP server of recipient. For example
 if you haven't any MTA available or want to check recipient's server responses (e.g. to know is
-such user exists on this server). First you need to know address of necessary SMTP server. We'll
-get it with help of L<Net::DNS>. Then we'll send it as usual
+such user exists on this server [see L<Mojo::Email::Checker::SMTP>]). First you need to know address
+of necessary SMTP server. We'll get it with help of L<Net::DNS>. Then we'll send it as usual
 
 	# will use non-blocking approach in this example
 	use strict;
