@@ -24,6 +24,12 @@ sub make_smtp_server {
 			while (my $resp = <$sock2>) {
 				syswrite($clt, $resp) && DEBUG && warn "<- $resp" if $resp =~ /^\d+/;
 				next if $resp =~ /^\d+-/;
+				if ($resp =~ /!quit\s*$/) {
+					warn "!quit\n" if DEBUG;
+					$clt->close();
+					last;
+				}
+				
 				my $cmd = <$clt> or last;
 				warn "-> $cmd" if DEBUG;
 				syswrite($sock2, $cmd);
